Attribute VB_Name = "modRPN"
Option Explicit

'This is the RPN (Reverse-Polish Notation) engine.  The RPN is an extremely fast
'and stable way to solve expressions.  Moving D++ over to RPN has increased D++'s
'capabilities dramatically.
'Normally an RPN engine is just used for math functions, however this one has been
'super-modified to accomodate for strings, variables, and functions.


'stuff used for independent testing of the RPN engine
'Sub main()
'
'End Sub

Public Function RPNcalculate(expression As String, Optional blockFunctions As Boolean = False) As String
    On Error GoTo err
    Dim aDebug As Boolean
    If Trim(expression) = "" Then Exit Function
    If debugMode Then
        If Left(expression, 1) = "@" Then
            expression = Mid(expression, 2): aDebug = True
            addDebug "RPN $> " & expression
        End If
    End If
    RPNcalculate = calculateRpn(convertToRpn(expression), aDebug, blockFunctions)

    Exit Function
err:
    'If noerror Then RPNcalculate = "Couldn't solve expression, probably invalid.": Exit Function
    FlagError "RPN Error", err.Description & vbCrLf & "Expression: " & expression, iLocation
    'FlagError "Run-Time Error", err.Description, iLocation
End Function

Public Function convertToRpn(ByVal expression As String) As String()
    'converts an expression to RPN
    
    Dim finalProduct() As String, tokenCount As Integer
    Dim tokens() As String, n As Long
    Dim tStack As New clsStack
    
    expression = "( " & Trim(expression) & " )"
    tokens = Tokenize(expression)
    'tokens = Split(expression, " ")
    'MsgBox displayArray(tokens, vbCrLf)
    ReDim finalProduct(UBound(tokens))
        
    tStack.ClearStack
    
    For n = LBound(tokens) To UBound(tokens)
        Select Case tokens(n)
            Case "(": tStack.Push "("
            Case ")"
                Do While tStack.Peek <> "("
                    finalProduct(tokenCount) = tStack.Pop
                    tokenCount = tokenCount + 1
                Loop
                If tStack.Peek = "(" Then tStack.Pop
            Case "+", "-", "*", "/", "\", "&", "^", "%", "<", ">", ">=", "<=", "<>", "!=", "=", "!", "&&", "||", "$", "|", "~"
                Do
                    If tStack.stackSize = 0 Or tStack.Peek = "(" Or GetPval(tStack.Peek) < GetPval(tokens(n)) Then
                        tStack.Push tokens(n)
                        Exit Do
                    Else
                        finalProduct(tokenCount) = tStack.Pop
                        tokenCount = tokenCount + 1
                    End If
                Loop
            Case Else
                finalProduct(tokenCount) = tokens(n)
                tokenCount = tokenCount + 1
        End Select
    Next n
    
    ReDim Preserve finalProduct(tokenCount - 1)
    
    convertToRpn = finalProduct
End Function

Public Function calculateRpn(expression() As String, aDebug As Boolean, blockFunctions As Boolean) As String
    'solves an RPN expression
    
    Dim tokens() As String, n As Long
    Dim temp As String
    Dim tStack As New clsStack
    
    'tokens = Split(Trim(displayArray(expression, " ")), Chr$(32))
    tokens = expression
    If aDebug Then addDebug "    $> " & displayArray(tokens, " ")
    
    tStack.ClearStack
    
    For n = LBound(tokens) To UBound(tokens)
        If IsOper(tokens(n), False) And tokens(n) <> "!" Then CheckStack tStack.stackSize
        Select Case tokens(n)
            Case "+": tStack.Push Val(tStack.Pop) + Val(tStack.Pop)
            Case "-": temp = tStack.Pop: tStack.Push Val(tStack.Pop) - Val(temp)
            Case "*": tStack.Push Val(tStack.Pop) * Val(tStack.Pop)
            Case "/": tStack.Push Val(tStack.Pop) / Val(tStack.Pop)
            Case "\": temp = tStack.Pop: tStack.Push Val(tStack.Pop) \ Val(temp)
            Case "^": tStack.Push Val(tStack.Pop) ^ Val(tStack.Pop)
            Case "%": tStack.Push Val(tStack.Pop) Mod Val(tStack.Pop)
            Case "&": temp = tStack.Pop: tStack.Push tStack.Pop & temp
            Case "!": tStack.Push IIf(Val(tStack.Pop) = 0, 1, 0)
            Case "<", ">", ">=", "<=", "<>", "!=", "=": temp = tStack.Pop: tStack.Push doEval(tStack.Pop, tokens(n), temp)
            Case "&&": tStack.Push IIf(tStack.Pop <> 0 And tStack.Pop <> 0, 1, 0)
            Case "||": tStack.Push IIf(tStack.Pop <> 0 Or tStack.Pop <> 0, 1, 0)
            Case Else
                If Left(tokens(n), 1) = Chr(34) Then 'quote
                    tStack.Push Mid(tokens(n), 2, Len(tokens(n)) - 2)
                ElseIf FindVar(tokens(n)) Then  'var
                    tStack.Push GetVar(tokens(n))
                ElseIf InStr(1, tokens(n), "(") <> 0 Then 'function
                    If Not blockFunctions Then
                        If Left(tokens(n), 3) = "rpn" Then
                            n = n
                        End If
                        tStack.Push GetFunctions(tokens(n))
                    Else
                        tStack.Push tokens(n)
                    End If
                Else 'k it's either a number or invalid, let someone else handle it :)
                    tStack.Push GetVar(tokens(n))
                End If
        End Select
    Next n
    
    If tStack.stackSize > 1 Then
        err.Raise 500, , "Invalid expression"
        End
    End If
    If aDebug Then addDebug "    $> " & tStack.Peek
    calculateRpn = tStack.Pop
End Function

Private Function CheckStack(tStackSize As Integer)
    Dim n As Long
    'before performing an operation, we must make sure the stack has enough stuff!
    'this prevents fatal subscript errors - new in 4.01
    If tStackSize < 2 Then
        'FlagError "Syntax Error", "Invalid expression.", iLocation
        'an error, but this will flag it up above
        n = 5 / 0
        End
    End If
End Function

Private Function GetPval(token As String) As Integer
    'these are VERY important to the RPN!  it determines which operations
    'to evaluate first.  the ones at the top are executed last and the ones
    'at the bottom are executed first.  it is NOT recommended that you change
    'these, unless you use some weird substandard math system
    Select Case token
        Case "||"
            GetPval = 1
        Case "&&"
            GetPval = 2
        Case "!=", "=", "<>"
            GetPval = 3
        Case ">", "<", ">=", "<="
            GetPval = 4
        Case "&"
            GetPval = 5
        Case "+", "-"
            GetPval = 6
        Case "*", "/", "\", "%"
            GetPval = 7
        Case "!"
            GetPval = 8
        Case "^"
            GetPval = 9
        Case "(", ")"
            GetPval = 10
        Case Else
            GetPval = 0
    End Select
End Function

Private Function Tokenize(ByVal expression As String) As String()
    'tokenizes an expression into an array
    Dim tokenArray() As String, tokenCount As Integer
    Dim i As Integer, j As Integer, n As Long
    Dim sChar As String
    Dim includeNegative As Boolean
    
    ReDim tokenArray(Len(expression))
    
    For i = 1 To Len(expression)
        sChar = Mid(expression, i, 1)
        
        If IsOper(sChar, True) Then   'operator
            If i < Len(expression) Then
                If IsOper(Mid(expression, i + 1, 1), False) Then _
                    sChar = Mid(expression, i, 2): i = i + 1
            End If
            If sChar = "-" Then
                If tokenCount = 1 Then
                    includeNegative = True: GoTo skip
                Else
                    If IsOper(tokenArray(tokenCount - 1), False) Then _
                        includeNegative = True: GoTo skip
                End If
            End If
            tokenArray(tokenCount) = sChar
            tokenCount = tokenCount + 1
        ElseIf IsNumeric(sChar) Then    'number
            j = i + 1
            Do Until j > Len(expression)
                If Not IsNumeric(Mid(expression, j, 1)) And Mid(expression, j, 1) <> "." Then Exit Do
                j = j + 1
            Loop
            tokenArray(tokenCount) = Mid(expression, i, j - i)
            If includeNegative Then tokenArray(tokenCount) = -1 * Val(tokenArray(tokenCount)): _
                includeNegative = False
            tokenCount = tokenCount + 1
            i = j - 1
        ElseIf sChar = Chr(34) Then 'quote: fixed string
            j = InStr(i + 1, expression, Chr(34))
            If j > 0 Then
                tokenArray(tokenCount) = Mid(expression, i, j - i + 1)
                tokenCount = tokenCount + 1
                i = j
            Else
                err.Raise 500, , "Expected end-quote"
                End
            End If
            If includeNegative Then includeNegative = False
        Else        'other: variable? function?
            If Not isValidChar(sChar) Then GoTo skip
            
            j = i + 1
            Do Until j > Len(expression)
                If Not isValidChar(Mid(expression, j, 1)) Then Exit Do
                sChar = Mid(expression, j, 1)
                If sChar = "(" Or sChar = "[" Then
                    j = FindEncloseEnd(j + 1, expression, sChar)
                    'an error, but this will flag it up above
                    If j = 0 Then err.Raise 500, , "Expected end-parenthesis"
                End If
                j = j + 1
            Loop
            tokenArray(tokenCount) = Mid(expression, i, j - i)
            tokenCount = tokenCount + 1
            i = j - 1
        End If
skip:
    Next 'i
    
    ReDim Preserve tokenArray(tokenCount - 1)
    Tokenize = tokenArray
End Function

Private Function IsOper(ByVal NumVal As String, Parens As Boolean) As Boolean
    'determines if a character is an operator
    Select Case NumVal
        Case "+", "-", "*", "/", "\", "&", "^", "%", "<", ">", "=", "!", "|"
            IsOper = True
        Case "(", ")"
            If Parens Then IsOper = True
    End Select
End Function

Public Function displayArray(theArray() As String, delimiter As String) As String
    Dim tempString As String, i As Integer
    
    For i = LBound(theArray) To UBound(theArray)
        tempString = tempString & theArray(i) & IIf(i <> UBound(theArray), delimiter, "")
    Next i
    
    displayArray = tempString
End Function

Private Function isValidChar(ByVal Char As String) As Boolean
    'determines if a character is valid
    Dim charCode As Integer
    charCode = Asc(Mid(Char, 1, 1))

    If charCode >= 65 And charCode <= 90 Then GoTo validChar      'upper case
    If charCode >= 97 And charCode <= 122 Then GoTo validChar     'lower case
    If charCode >= 48 And charCode <= 57 Then GoTo validChar      'numbers
    If charCode = 95 Or charCode = 46 Then GoTo validChar        '_ & .
    If charCode = 91 Or charCode = 93 Then GoTo validChar        ' [ ]
    If charCode = 40 Or charCode = 41 Then GoTo validChar        ' ( )
    'If charCode = 33 Then GoTo validChar
    
    Exit Function
    
validChar:
    isValidChar = True
End Function

'more stuff for independent training
'Public Function GetVar(varname As String) As String
'    If IsNumeric(varname) Then GetVar = varname: Exit Function
'
'    Select Case varname
'        Case "var1"
'            GetVar = "variable #1"
'        Case "var2"
'            GetVar = 238
'        Case "var3"
'            GetVar = "var: 238"
'        Case "crlf"
'            GetVar = vbCrLf
'        Case Else
'            MsgBox "undefined var: " & varname
'            End
'    End Select
'End Function

